/*	set_lib.c

	The following is a general-purpose set library which
	allows C programs to be written nearly as though the
	set data-type existed... sets are structs which can
	be freely copied by assignment statements, passed by
	value to functions, returned by value from functions,
	and allocated by ordinary declarations.

	Inconveniences:
	1. Sets are of fixed size because C cannot track a
	   runtime-sized array (no paramtypes yet).
	2. Sets cannot be compared without a function call.
	3. Some strangeness happens if only a portion of a
	   set is used or if there are bits in a word which
	   are ignored in the set representation; e.g., the
	   set_not routine may not function correctly.

	(C) 1987 by Hank Dietz
*/

#include "sets.h"

unsigned bitmask[32] = {
	0x00000001, 0x00000002, 0x00000004, 0x00000008,
	0x00000010, 0x00000020, 0x00000040, 0x00000080,
	0x00000100, 0x00000200, 0x00000400, 0x00000800,
	0x00001000, 0x00002000, 0x00004000, 0x00008000,
	0x00010000, 0x00020000, 0x00040000, 0x00080000,
	0x00100000, 0x00200000, 0x00400000, 0x00800000,
	0x01000000, 0x02000000, 0x04000000, 0x08000000,
	0x10000000, 0x20000000, 0x40000000, 0x80000000
};

int
set_deg(a)
set a;
{
	/* Fast compute degree of a set... the number
	   of elements present in the set.  Assumes
	   that all word bits are used in the set
	   and that SETSIZE is a multiple of WORDSIZE.
	*/
	register unsigned *p = &(a.setword[0]);
	register unsigned *endp = &(a.setword[SETWORDS]);
	register degree = 0;

	do {
		register unsigned t = *p;
		register unsigned *b = &(bitmask[0]);
		do {
			if (t & *b) ++degree;
		} while (++b < &(bitmask[WORDSIZE]));
	} while (++p < endp);

	return(degree);
}

set
set_or(b,c)
set b, c;
{
	/* Fast set union operation */
	register unsigned *p = &(b.setword[0]);
	register unsigned *q = &(c.setword[0]);
	register unsigned *endp = &(b.setword[SETWORDS]);

	do {
		*p |= *(q++);
	} while (++p < endp);
	return(b);
}

set
set_and(b,c)
set b, c;
{
	/* Fast set intersection operation */
	register unsigned *p = &(b.setword[0]);
	register unsigned *q = &(c.setword[0]);
	register unsigned *endp = &(b.setword[SETWORDS]);

	do {
		*p &= *(q++);
	} while (++p < endp);
	return(b);
}

set
set_dif(b,c)
set b, c;
{
	/* Fast set difference operation */
	register unsigned *p = &(b.setword[0]);
	register unsigned *q = &(c.setword[0]);
	register unsigned *endp = &(b.setword[SETWORDS]);

	do {
		*p &= (~ *(q++));
	} while (++p < endp);
	return(b);
}

set
set_of(b)
int b;
{
	/* Fast singleton set constructor operation */
	static set a;
	register unsigned *p = &(a.setword[0]);
	register unsigned *endp = &(a.setword[SETWORDS]);

	do {
		*p = 0;
	} while (++p < endp);
	if (b != nil) {
		a.setword[DIVWORD(b)] = bitmask[MODWORD(b)];
	}
	return(a);
}

set
set_not(a)
set a;
{
	/* Fast not of set a (assumes all bits used) */
	register unsigned *p = &(a.setword[0]);
	register unsigned *endp = &(a.setword[SETWORDS]);

	do {
		*p = (~ *p);
	} while (++p < endp);
	return(a);
}

int
set_equ(a,b)
set a, b;
{
	/* Fast set equality comparison operation */
	register unsigned *p = &(a.setword[0]);
	register unsigned *q = &(b.setword[0]);
	register unsigned *endp = &(a.setword[SETWORDS]);

	do {
		if (*p != *q) return(0);
		++q;
	} while (++p < endp);
	return(1);
}

int
set_sub(a,b)
set a, b;
{
	/* Fast check for a is a proper subset of b (alias a < b) */
	register unsigned *p = &(a.setword[0]);
	register unsigned *q = &(b.setword[0]);
	register unsigned *endp = &(a.setword[SETWORDS]);
	register int asubset = 0;

	do {
		/* Prune tests based on guess that most set words
		   will match, particularly if a is a subset of b.
		*/
		if (*p != *q) {
			if (*p & ~(*q)) {
				/* Fail -- a contains something b does not */
				return(0);
			}
			/* At least this word was a proper subset, hence
			   even if all other words are equal, a is a
			   proper subset of b.
			*/
			asubset = 1;
		}
		++q;
	} while (++p < endp);

	return(asubset);
}

int
set_int(b)
set b;
{
	/* Fast pick any element of the set b */
	register unsigned *p = &(b.setword[0]);
	register unsigned *endp = &(b.setword[SETWORDS]);

	do {
		if (*p) {
			/* Found a non-empty word of the set */
			register int i = ((p - &(b.setword[0])) * WORDSIZE);
			register unsigned t = *p;
			p = &(bitmask[0]);
			while (!(*p & t)) {
				++i; ++p;
			}
			return(i);
		}
	} while (++p < endp);

	/* Empty -- only element it contains is nil */
	return(nil);
}

set_el(b,a)
int b;
set a;
{
	/* nil is an element of every set */
	if (b == nil) return(1);

	/* Otherwise, we have to check */
	return(a.setword[DIVWORD(b)] & bitmask[MODWORD(b)]);
}

set_nil(a)
set a;
{
	/* Fast check for nil set */
	register unsigned *p = &(a.setword[0]);
	register unsigned *endp = &(a.setword[SETWORDS]);

	/* The set is not empty if any word used to store
	   the set is non-zero.  This means one must be a
	   bit careful about doing things like negation.
	*/
	do {
		if (*p) return(0);
	} while (++p < endp);
	return(1);
}

char *
set_str(a)
set a;
{
	/* Fast convert set a into ASCII char string...
	   assumes that all word bits are used in the set
	   and that SETSIZE is a multiple of WORDSIZE.
	   Trailing 0 bits are removed from the string.
	*/
	register unsigned *p = &(a.setword[0]);
	register unsigned *endp = &(a.setword[SETWORDS]);
	static char str_tmp[SETSIZE+1];
	register char *q = &(str_tmp[0]);

	do {
		register unsigned t = *p;
		register unsigned *b = &(bitmask[0]);
		do {
			*(q++) = ((t & *b) ? '1' : '0');
		} while (++b < &(bitmask[WORDSIZE]));
	} while (++p < endp);

	/* Trim trailing 0s & NULL terminate the string */
	while ((q > &(str_tmp[1])) && (*(q-1) != '0')) --q;
	*q = 0;

	return(&(str_tmp[0]));
}

set
set_val(s)
register char *s;
{
	/* Fast convert set ASCII char string into a set.
	   If the string ends early, the remaining set bits
	   are all made zero.
	*/
	static set a;
	register unsigned *p = &(a.setword[0]);
	register unsigned *endp = &(a.setword[SETWORDS]);

	do {
		register unsigned *b = &(bitmask[0]);
		/* Start with a word with no bits on */
		*p = 0;
		do {
			if (*s) {
				if (*s == '1') {
					/* Turn-on this bit */
					*p |= *b;
				}
				++s;
			}
		} while (++b < &(bitmask[WORDSIZE]));
	} while (++p < endp);

	return(a);
}

int
set_hash(a, mod)
set a;
register int mod;
{
	/* Fast hash of set a (assumes all bits used) */
	register unsigned *p = &(a.setword[0]);
	register unsigned *endp = &(a.setword[SETWORDS]);
	register int i = 0;

	do {
		i = ((i + i + (i<0)) ^ (~ *p));
	} while (++p < endp);

	return(((i< 0) ? -i : i) % mod);
}
